@@ -13,7 +13,7 @@ from fabric import Connection, task
 from invoke import Collection  # noqa, re-export
 
 
-__version__ = "1.0.20200924"
+__version__ = "1.0.20201221"
 
 
 # I don't care, in this context.
@@ -58,7 +58,14 @@ class Config:
         pre_commit_hook()
 
     def __getattr__(self, key):
-        terminate(f"Configuration key '{key}' not set")
+        environments = self.__dict__.get("environments", "")
+        if environments:
+            environments = f" [{', '.join(sorted(environments))}]"
+        terminate(
+            f"Configuration key '{key}' not set. "
+            f"Maybe you forgot to set an environment with which to interact?"
+            f"{environments}"
+        )
 
 
 #: Defaults
@@ -73,10 +80,10 @@ class Connection(Connection):
         super().__init__(*args, **kwargs)
 
 
-def pre_commit_hook():
+def pre_commit_hook(*, force=False):
     """Install the pre-commit hook running coding style checks"""
     path = config.base / ".git" / "hooks" / "pre-commit"
-    if not path.exists():
+    if not path.exists() or force:
         with path.open("w") as hook:
             hook.write("#!/bin/sh\nfl check\n")
         path.chmod(0o755)
@@ -120,6 +127,12 @@ for job in $(jobs -p); do wait $job; done
         run(ctx, f"bash {f.name}", replace_env=False)
 
 
+@task
+def hook(ctx):
+    """Install the pre-commit hook"""
+    pre_commit_hook(force=True)
+
+
 @task(auto_shortflags=False)
 def dev(ctx, host="127.0.0.1", port=8000):
     """Run the development server for the frontend and backend"""
@@ -128,23 +141,11 @@ def dev(ctx, host="127.0.0.1", port=8000):
         ctx,
         [
             f"venv/bin/python manage.py runserver 0.0.0.0:{port}",
-            f'HOST="{host}" node_modules/.bin/webpack-dev-server'
-            f" --host 0.0.0.0 --port 4000 --hot",
+            f'HOST="{host}" npx webpack-dev-server --host 0.0.0.0 --port 4000 --hot',
         ],
     )
 
 
-def _reset_passwords(ctx):
-    # 'password' encoded with a constant salt. Does not force a login after pull_db
-    pw = r"pbkdf2_sha256\$216000\$salt\$xuFh/Jmp9ZyNeO4k67igyjH9t5hHZ84M69rSfrV2W/g="
-    run(
-        ctx,
-        f"venv/bin/python manage.py shell -c \"pw='{pw}';"
-        f"from django.contrib.auth import get_user_model as g;"
-        f'g()._base_manager.update(password=pw)"',
-    )
-
-
 @task
 def pull_db(ctx):
     """Pull a local copy of the remote DB and reset all passwords"""
@@ -159,7 +160,20 @@ def pull_db(ctx):
     run(ctx, f"createdb {dbname}")
     run(ctx, f"ssh {config.host} -C 'pg_dump -Ox {srv_dsn}' | psql {local_dsn}")
 
-    _reset_passwords(ctx)
+    reset_pw(ctx)
+
+
+@task
+def reset_pw(ctx):
+    """Set all user passwords to ``"password"``"""
+    # 'password' encoded with a constant salt. Does not force a login after pull_db
+    pw = r"pbkdf2_sha256\$216000\$salt\$xuFh/Jmp9ZyNeO4k67igyjH9t5hHZ84M69rSfrV2W/g="
+    run(
+        ctx,
+        f"venv/bin/python manage.py shell -c \"pw='{pw}';"
+        f"from django.contrib.auth import get_user_model as g;"
+        f'g()._base_manager.update(password=pw)"',
+    )
 
 
 def _local_env(path=".env"):
@@ -179,10 +193,6 @@ def _srv_env(conn, path):
     return lambda *a, **kw: speckenv.env(*a, **kw, mapping=mapping)
 
 
-def _nine_has_manage_databases(conn):
-    return bool(run(conn, "which nine-manage-databases").stdout.strip())
-
-
 @task
 def mm(ctx):
     """Update the translation catalogs"""
@@ -217,12 +227,16 @@ def _python3():
     return next(filter(None, (shutil.which(v) for v in interpreters)))
 
 
+def _pip_up(c):
+    run(c, "venv/bin/python -m pip install -U 'pip!=20.3.2' wheel setuptools")
+
+
 @task(auto_shortflags=False, help={"stable": "Avoid pre-release versions of packages"})
 def upgrade(ctx, stable=False):
     """Re-create the virtualenv with newest versions of all libraries"""
     run(ctx, "rm -rf venv")
     run(ctx, f"{_python3()} -m venv venv")
-    run(ctx, "venv/bin/python -m pip install -U pip wheel setuptools")
+    _pip_up(ctx)
     extra = "" if stable else "--pre"
     run(ctx, f"venv/bin/python -m pip install -U -r requirements-to-freeze.txt {extra}")
     freeze(ctx)
@@ -245,7 +259,7 @@ def update(ctx):
     venv = config.base / "venv"
     if not venv.exists():
         run(ctx, f"{_python3()} -m venv venv")
-    run(ctx, "venv/bin/python -m pip install -U pip wheel setuptools")
+    _pip_up(ctx)
     run(ctx, "venv/bin/python -m pip install -r requirements.txt")
     run(ctx, 'find . -name "*.pyc" -delete')
     run(ctx, "yarn")
@@ -263,12 +277,13 @@ def _local_dotenv_if_not_exists():
     with dotenv.open("w") as f:
         f.write(
             f"""\
+DEBUG=True
 DATABASE_URL=postgres://localhost:5432/{dbname}
 CACHE_URL=hiredis://localhost:6379/1/?key_prefix={dbname}
 SECRET_KEY={secret_key}
-SENTRY_DSN=
 ALLOWED_HOSTS=["*"]
-DEBUG=True
+
+SENTRY_DSN=
 """
         )
 
@@ -295,10 +310,10 @@ def nine_vhost(ctx):
             conn,
             f"sudo nine-manage-vhosts virtual-host create {config.domain}"
             " --template=feinheit_cache"
-            " --webroot=/home/www-data/{config.domain}/htdocs",
+            f" --webroot=/home/www-data/{config.domain}/htdocs",
         )
         with conn.cd(config.domain):
-            run(conn, "mkdir -f media tmp")
+            run(conn, "mkdir -p media tmp")
 
 
 @task(auto_shortflags=False, help={"include-www": "Include the www. subdomain"})
@@ -345,6 +360,10 @@ def nine_unit(ctx):
         run(conn, f"systemctl --user enable gunicorn@{config.domain}.service")
 
 
+def _nine_has_manage_databases(conn):
+    return bool(run(conn, "which nine-manage-databases", warn=True).stdout.strip())
+
+
 @task
 def nine_db_dotenv(ctx):
     """Create a database and initialize the .env"""
@@ -371,14 +390,17 @@ def nine_db_dotenv(ctx):
         else:
             run(
                 conn,
-                f'psql -c "CREATE ROLE {dbname} WITH'
+                f'source ~/.profile && psql -c "CREATE ROLE {dbname} WITH'
                 f" ENCRYPTED PASSWORD '{password}'"
                 f' LOGIN NOCREATEDB NOCREATEROLE NOSUPERUSER"',
             )
-            run(conn, f'psql -c "GRANT {dbname} TO admin"')
             run(
                 conn,
-                f'psql -c "CREATE DATABASE {dbname} WITH'
+                f'source ~/.profile && psql -c "GRANT {dbname} TO admin"',
+            )
+            run(
+                conn,
+                f'source ~/.profile && psql -c "CREATE DATABASE {dbname} WITH'
                 f" OWNER {dbname} TEMPLATE template0 ENCODING 'UTF8'"
                 f'"',
             )
@@ -390,10 +412,11 @@ DEBUG=False
 DATABASE_URL=postgres://{dbname}:{password}@localhost:5432/{dbname}
 CACHE_URL=hiredis://localhost:6379/1/?key_prefix={dbname}
 SECRET_KEY={secret_key}
-SENTRY_DSN=
 ALLOWED_HOSTS=[".{config.domain}", ".{conn.host}"]
+
 GOOGLE_CLIENT_ID=
 GOOGLE_CLIENT_SECRET=
+SENTRY_DSN=
 
 # LIVE=True
 # CANONICAL_DOMAIN={config.domain}
@@ -420,6 +443,15 @@ def nine_ssl(ctx):
         )
 
 
+@task
+def nine_restart(ctx):
+    """Restart the application server"""
+    with Connection(config.host) as conn:
+        with conn.cd(config.domain):
+            run(conn, "venv/bin/python manage.py check --deploy")
+        run(conn, f"systemctl --user restart gunicorn@{config.domain}.service")
+
+
 @task
 def nine_disable(ctx):
     """Disable a virtual host, dump and remove the DB and stop the gunicorn@ unit"""
@@ -427,22 +459,24 @@ def nine_disable(ctx):
         run(conn, f"sudo nine-manage-vhosts virtual-host remove {config.domain}")
         run(conn, f"systemctl --user stop gunicorn@{config.domain}.service")
         run(conn, f"systemctl --user disable gunicorn@{config.domain}.service")
+
         e = _srv_env(conn, f"{config.domain}/.env")
         srv_dsn = e("DATABASE_URL")
-        run(conn, f"pg_dump -Ox {srv_dsn} > DUMP.sql")
-        srv_dbname = _dbname_from_dsn(srv_dsn)
+        srv_dbname = _dbname_from_dsn(e("DATABASE_URL"))
+
+        run(conn, f"pg_dump -Ox {srv_dsn} > {config.domain}/{srv_dbname}.sql")
 
         if _nine_has_manage_databases(conn):
             run(conn, f"sudo nine-manage-databases database drop --force {srv_dbname}")
         else:
-            run(conn, f"dropdb {srv_dbname}")
-            run(conn, f"dropuser {srv_dbname}")
+            run(conn, f"source ~/.profile && dropdb {srv_dbname}")
+            run(conn, f"source ~/.profile && dropuser {srv_dbname}")
 
 
 @task
 def nine_checkout(ctx):
     """Checkout the repository on the server"""
-    repo = run(ctx, "git config remote.origin.url", hide=True).stdout
+    repo = run(ctx, "git config remote.origin.url", hide=True).stdout.strip()
     with Connection(config.host) as conn:
         run(conn, f"git clone {repo} {config.domain} -b {config.branch}")
 
@@ -454,7 +488,7 @@ def nine_venv(ctx):
         with conn.cd(config.domain):
             run(conn, "rm -rf venv")
             run(conn, "PATH=~/.pyenv/shims:$PATH python3 -m venv venv")
-            run(conn, "venv/bin/python -m pip install -U pip wheel setuptools")
+            _pip_up(conn)
             run(conn, "venv/bin/python -m pip install -r requirements.txt")
 
 
@@ -474,14 +508,19 @@ def nine(ctx):
 def bitbucket(ctx):
     """Create a repository on Bitbucket and push the code"""
     e = _local_env(Path.home() / ".box.env")
-    print("Username: ", end="")
-    username = input(e("BITBUCKET_USERNAME"))
-    print("Password: ", end="")
-    password = input(e("BITBUCKET_PASSWORD"))
-    print("Organization: ", end="")
-    organization = input(e("BITBUCKET_ORGANIZATION"))
-    print("Repository: ", end="")
-    repository = input(config.domain)
+    username = e("BITBUCKET_USERNAME")
+    password = e("BITBUCKET_PASSWORD")
+    organization = e("BITBUCKET_ORGANIZATION")
+    repository = config.domain
+
+    print(f"Username [{username}]: ", end="")
+    username = input() or username
+    print(f"Password [{password}]: ", end="")
+    password = input() or password
+    print(f"Organization [{organization}]: ", end="")
+    organization = input() or organization
+    print(f"Repository [{repository}]: ", end="")
+    repository = input() or repository
 
     run(
         ctx,
@@ -495,6 +534,26 @@ curl -X POST -v -u {username}:"{password}" -H "content-type: application/json"\
     run(ctx, f"git push -u origin {config.branch}")
 
 
+@task
+def github(ctx):
+    """Create a repository on GitHub and push the code"""
+    url = run(ctx, "git config remote.origin.url", hide=True, warn=True).stdout.strip()
+    if url:
+        terminate(f"The 'origin' remote already points to '{url}'")
+
+    e = _local_env(Path.home() / ".box.env")
+    organization = e("GITHUB_ORGANIZATION")
+    repository = config.domain
+
+    print(f"Organization [{organization}]: ", end="")
+    organization = input() or organization
+    print(f"Repository [{repository}]: ", end="")
+    repository = input() or repository
+
+    run(ctx, f"gh repo create {organization}/{repository} --private")
+    run(ctx, f"git push -u origin {config.branch}")
+
+
 @task
 def fetch(ctx):
     """Ensure a remote exists for the server and fetch"""
@@ -518,13 +577,13 @@ def _check_django(ctx):
 def _check_prettier(ctx):
     run(
         ctx,
-        f"node_modules/.bin/prettier --list-different --no-semi"
+        f"npx prettier --list-different --no-semi"
         f' "*.js" "{config.app}/static/**/*.js" "{config.app}/static/**/*.scss"',
     )
 
 
 def _check_eslint(ctx):
-    run(ctx, f'node_modules/.bin/eslint "*.js" {config.app}/static')
+    run(ctx, f'npx eslint "*.js" {config.app}/static')
 
 
 def _check_branch(ctx):
@@ -533,6 +592,14 @@ def _check_branch(ctx):
         terminate(f"Current branch is '{branch}', should be '{config.branch}'")
 
 
+def _check_no_uncommitted_changes(ctx):
+    with Connection(config.host) as conn:
+        with conn.cd(config.domain):
+            result = run(conn, "git status --porcelain").stdout.strip()
+            if result:
+                terminate("Terminating because of uncommitted changes on server")
+
+
 @task
 def check(ctx):
     """Check the coding style"""
@@ -545,7 +612,7 @@ def check(ctx):
 def _fmt_prettier(ctx):
     run(
         ctx,
-        f"node_modules/.bin/prettier --write --no-semi"
+        f"npx prettier --write --no-semi"
         f' "*.js" "{config.app}/static/**/*.js" "{config.app}/static/**/*.scss"',
     )
 
@@ -575,13 +642,15 @@ def fmt(ctx):
     _fmt_prettier(ctx)
 
 
-@task
-def deploy(ctx):
+@task(auto_shortflags=False, help={"fast": "Skip the Webpack build"})
+def deploy(ctx, fast=False):
     """Deploy once 🔥"""
     _check_branch(ctx)
+    _check_no_uncommitted_changes(ctx)
     check(ctx)
     run(ctx, f"git push origin {config.branch}")
-    run(ctx, "NODE_ENV=production node_modules/.bin/webpack -p --bail")
+    if not fast:
+        run(ctx, "NODE_ENV=production npx webpack -p --bail")
 
     with Connection(config.host) as conn:
         with conn.cd(config.domain):
@@ -589,15 +658,16 @@ def deploy(ctx):
             run(conn, "git fetch origin")
             run(conn, f"git merge --ff-only origin/{config.branch}")
             run(conn, 'find . -name "*.pyc" -delete')
-            run(conn, "venv/bin/python -m pip install -U pip wheel setuptools")
+            _pip_up(conn)
             run(conn, "venv/bin/python -m pip install -r requirements.txt")
             run(conn, "venv/bin/python manage.py migrate")
             run(conn, "venv/bin/python manage.py check --deploy", warn=True)
-            run(
-                ctx,
-                f"rsync -pthrz --delete --stats"
-                f" static/ {config.host}:{config.domain}/static/",
-            )
+            if not fast:
+                run(
+                    ctx,
+                    f"rsync -pthrz --delete --stats"
+                    f" static/ {config.host}:{config.domain}/static/",
+                )
             run(conn, "venv/bin/python manage.py collectstatic --noinput")
         run(conn, f"systemctl --user restart gunicorn@{config.domain}.service")
 
@@ -605,6 +675,7 @@ def deploy(ctx):
 
 
 GENERAL = {
+    hook,
     cm,
     dev,
     mm,
@@ -612,8 +683,10 @@ GENERAL = {
     freeze,
     update,
     pull_db,
+    reset_pw,
     local,
     bitbucket,
+    github,
     fetch,
     check,
     fmt,
@@ -626,6 +699,7 @@ NINE = {
     nine_unit,
     nine_db_dotenv,
     nine_ssl,
+    nine_restart,
     nine_disable,
     nine_checkout,
     nine_venv,
